#!/usr/bin/perl

sub read_splits {

    local($prefix, *a, *b) = @_;

    open(SPLITS, "$prefix\_splits");
    $splits = join("", <SPLITS>); 

    while($splits =~ m/(.*)\n/g) {

	$line = $1; 

	if($line =~ m/\{(.*)\}\s*\{(.*)\}/) {

	    push @a, $1; push @b, $2;
	}
	else {

	    print "Problem with $line\n";
	}
    }

    close(SPLITS);
}

sub intersection_and_complement {

    local($a, $b) = @_;

    local(@a) = (sort(split(/\s+/, $a)));
    local(@b) = (sort(split(/\s+/, $b)));
    local(@intersection, @complement) = ();

    $a_index = $b_index = 0;

    while($a_index <= $#a) {

	while($a[$a_index] gt $b[$b_index] && $b_index < $#b) { $b_index++; }

	if($a[$a_index] eq $b[$b_index]) { push @intersection, $a[$a_index]; }
	else { push @complement, $a[$a_index]; }

	$a_index++;
    }

    return (join(" ", @intersection), join(" ", @complement));
}

sub get_bin_pval {

    local($count, $n, $freq) = @_;
    local($pval) = (1);

    if($count/$n > $freq) {

	$pval = `./binomial_invcdf $count $n $freq`; chomp $pval; 
    }

    return $pval;
}
	
sub compare_sets {

    local($a, $b, *pair_ge, *pair_le, $ge_freq, $le_freq) = @_;

    local(@a) = (sort(split(/\s+/, $a)));
    local(@b) = (sort(split(/\s+/, $b)));
    local($ge_count, $le_count) = (0, 0);

    foreach $taxa1 (@a) {

	foreach $taxa2 (@b) {

	    $pair = join(" ", ($taxa1, $taxa2));
	    $ge_count += $pair_ge{$pair}; $le_count += $pair_le{$pair};
	}
    }
    
    local($n) = (scalar(@a)*scalar(@b));

    $ge_pval = &get_bin_pval($ge_count, $n, $ge_freq);
    $le_pval = &get_bin_pval($le_count, $n, $le_freq);
    
    print LOG "($ge_count, $n, $ge_freq, $ge_pval ,,, $le_count, $n, $le_freq, $le_pval)\n";

    return ($ge_pval, $le_pval);
}

sub test_candidate {

    local($candidate, $o1, $o2, $o3, *pair_ge, *pair_le, $ge_freq, $le_freq, $num_of_bin_tests, $evalue_threshold) = @_;
    local($greater, $lesser, $error) = (0, 0, 0);

    foreach $other ($o1, $o2, $o3) {

	@taxa = split(/\s+/, $other); 

	local($ge_pval, $le_pval) = &compare_sets($candidate, $other, *pair_ge, *pair_le, $ge_freq, $le_freq);
	local($g) = ($ge_pval < 0.01 ? 1 : 0);
	local($l) = ($le_pval < 0.01 ? 1 : 0);

	$greater += ($g == 1 ? ($l == 1 ? 0.5 : 1) : 0);
	$lesser += ($l == 1 ? ($g == 1 ? 0.5 : 1) : 0);
    }

    printf LOG "Test results: $greater $lesser %s\n", ($greater*$lesser >= 0.5 ? "***" : "");
    return ($greater*$lesser >= 0.5);
}

#
# $selective = 0 => Include all labels
#              1 => Only distance test
#

($dir, $selective) = @ARGV;
$graph_file = "$dir/graph";
$label_file = "$dir/graph.labels";

$num_of_bin_tests = `wc $graph_file`; chomp $num_of_bin_tests; $num_of_bin_tests =~ s/(\S+)/\1/;
$num_of_bin_tests *= 6;

@left_a = @left_b = ();
&read_splits("$dir/left", *left_a, *left_b);

@right_a = @right_b = ();
&read_splits("$dir/right", *right_a, *right_b);

if($selective) {

    open(PAIR_TEST, "$dir/pair_test_results");
    $num_of_mw_tests = `wc $dir/pair_test_results`; chomp $num_of_mw_tests; $num_of_mw_tests =~ s/\s+(\S+).*/\1/;
    $ge_count = 0; $le_count = 0; $evalue_threshold = 0.01;
    %pair_ge = (); %pair_le = ();

    while(<PAIR_TEST>) {
	
	($taxa1, $taxa2, $is_greater, $log_pvalue) = split;

	$pair = "$taxa1 $taxa2"; $inv_pair = "$taxa2 $taxa1";
	$pair_ge{$inv_pair} = $pair_ge{$pair} = ($is_greater ? (exp($log_pvalue)*$num_of_mw_tests < $evalue_threshold ? 1 : 0) : 0);
	$pair_le{$inv_pair} = $pair_le{$pair} = (!$is_greater ? (exp($log_pvalue)*$num_of_mw_tests < $evalue_threshold ? 1 : 0) : 0);
	$ge_count += $pair_ge{$pair}; $le_count += $pair_le{$pair};
    }
    close(PAIR_TEST);

    $ge_freq = $ge_count/$num_of_mw_tests; $le_freq = $le_count/$num_of_mw_tests;
}

%set_index = (); $curr_index = 0;

open(LABEL, ">$label_file");
open(GRAPH, $graph_file);
open(OUT, ">$dir/graph.labelled");
open(LOG, ">$dir/graph.labels.log");
$outfile = OUT;

while(<GRAPH>) {

    ($left_split, $right_split) = split;

    print $outfile "$left_split $right_split ";

    @result = &intersection_and_complement($left_a[$left_split], $right_a[$right_split]);

    (print LABEL "$curr_index $result[0]\n\n"), $set_index{$result[0]} = $curr_index++ if($set_index{$result[0]} eq "");
    (print LABEL "$curr_index $result[1]\n\n"), $set_index{$result[1]} = $curr_index++ if($set_index{$result[1]} eq "");

    print "Trouble: $left_split,$right_split = $left_a[$left_split],,,$right_a[$right_split]\n" if($result[0] eq "" || $result[1] eq "");
    $a = $result[0]; $b = $result[1];

    @result = &intersection_and_complement($left_b[$left_split], $right_a[$right_split]);

    (print LABEL "$curr_index $result[0]\n\n"), $set_index{$result[0]} = $curr_index++ if($set_index{$result[0]} eq "");
    (print LABEL "$curr_index $result[1]\n\n"), $set_index{$result[1]} = $curr_index++ if($set_index{$result[1]} eq "");

    print "Trouble: $left_split,$right_split = $left_b[$left_split],,,$right_a[$right_split]\n" if($result[0] eq "" || $result[1] eq "");
    $c = $result[0]; $d = $result[1];

    @set_a = split(/\s+/, $a); @set_b = split(/\s+/, $b); @set_c = split(/\s+/, $c); @set_d = split(/\s+/, $d);
    $size{$a} = scalar(@set_a); $size{$b} = scalar(@set_b); $size{$c} = scalar(@set_c); $size{$d} = scalar(@set_d);
    ($a, $b, $c, $d) = sort {$size{$a} <=> $size{$b}} ($a, $b, $c, $d);

    if(!$selective) {

	print $outfile "$set_index{$a} $set_index{$b} $set_index{$c}\n"; 
	next;
    }

    print LOG "Splits: $left_split $right_split\n";

    print LOG "Candidate: $set_index{$a}\n";    
    print $outfile "$set_index{$a} " if(&test_candidate($a, $b, $c, $d, *pair_ge, *pair_le, $ge_freq, $le_freq, $num_of_bin_tests, $evalue_threshold));

    print LOG "Candidate: $set_index{$b}\n";    
    print $outfile "$set_index{$b} " if(&test_candidate($b, $a, $c, $d, *pair_ge, *pair_le, $ge_freq, $le_freq, $num_of_bin_tests, $evalue_threshold));

    print LOG "Candidate: $set_index{$c}\n";    
    print $outfile "$set_index{$c} " if(&test_candidate($c, $a, $b, $d, *pair_ge, *pair_le, $ge_freq, $le_freq, $num_of_bin_tests, $evalue_threshold));

    print LOG (('-'x80)."\n");
    print $outfile "\n";
}

close(LOG);
close(GRAPH);
close(LABEL);
